/*
   AtrMemoryImage.C - access ATR images in RAM.

   Copyright (C) 2002 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "AtrMemoryImage.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_ZLIB
#include <zlib.h>
#endif

#include <ctype.h>

#include "DCMCodec.h"
#include "AtariDebug.h"

AtrMemoryImage::AtrMemoryImage()
	: fData(0), fChanged(false)
{
}

AtrMemoryImage::~AtrMemoryImage()
{
	FreeImageData();
}

void AtrMemoryImage::FreeImageData()
{
	if (fData) {
		delete[] fData;
		fData=0;
	}
	SetFormat(eNoDisk);
}

bool AtrMemoryImage::CreateImage(EDiskFormat format)
{
	size_t imgSize;
	FreeImageData();
	fChanged = true;
	if (!SetFormat(format)) {
		DPRINTF("SetFormat failed\n");
		return false;
	}

	SetWriteProtect(false);

	if ((imgSize = GetImageSize()) == 0) {
		DPRINTF("GetImageSize = 0\n");
		return false;
	}
	fData = new unsigned char[imgSize];

	if (fData) {
		memset (fData, 0, imgSize);
		return true;
	} else {
		DPRINTF("cannot alloc fData\n");
		FreeImageData();
		return false;
	}
}

bool AtrMemoryImage::CreateImage(ESectorLength density, size_t sectors)
{
	size_t imgSize;
	FreeImageData();
	fChanged = true;
	if (!SetFormat(density, sectors)) {
		DPRINTF("SetFormat failed\n");
		return false;
	}

	SetWriteProtect(false);

	if ((imgSize = GetImageSize()) == 0) {
		DPRINTF("GetImageSize = 0\n");
		return false;
	}
	fData = new unsigned char[imgSize];

	if (fData) {
		memset (fData, 0, imgSize);
		return true;
	} else {
		DPRINTF("cannot alloc fData\n");
		FreeImageData();
		return false;
	}
}

bool AtrMemoryImage::CreateImage(ESectorLength density, size_t sectorsPerTrack, size_t tracks, size_t sides)
{
	size_t imgSize;
	FreeImageData();
	fChanged = true;
	if (!SetFormat(density, sectorsPerTrack, tracks, sides)) {
		DPRINTF("SetFormat failed\n");
		return false;
	}

	SetWriteProtect(false);

	if ((imgSize = GetImageSize()) == 0) {
		DPRINTF("GetImageSize = 0\n");
		return false;
	}
	fData = new unsigned char[imgSize];

	if (fData) {
		memset (fData, 0, imgSize);
		return true;
	} else {
		DPRINTF("cannot alloc fData\n");
		FreeImageData();
		return false;
	}
}


bool AtrMemoryImage::ReadImageFromFile(const char* filename)
{
	bool ret;
	EImageType imageType = DetermineImageTypeFromFilename(filename);

	if (imageType == eUnknownImageType) {
		DPRINTF("cannot determine image type from filename - using ATR format\n");
		imageType = eAtrImageType;
	}

#ifndef USE_ZLIB
	switch (imageType) {
	case eAtrGzImageType:
	case eXfdGzImageType:
	case eDcmGzImageType:
	case eDiGzImageType:
		DPRINTF("cannot read compressed file - zlib support disabled at compiletime!\n");
		return false;
	default: break;
	}
#endif

	switch (imageType) {
	case eAtrImageType:
	case eAtrGzImageType:
		ret = ReadImageFromAtrFile(filename);
		break;
	case eXfdImageType:
	case eXfdGzImageType:
		ret = ReadImageFromXfdFile(filename);
		break;
	case eDcmImageType:
	case eDcmGzImageType:
		ret = ReadImageFromDcmFile(filename);
		break;
	case eDiImageType:
	case eDiGzImageType:
		ret = ReadImageFromDiFile(filename);
		break;
	case eUnknownImageType:
		DPRINTF("unknown image type!\n");
		return false;
	default:
		DPRINTF("unsupported image type!\n");
		return false;
	}
	return ret;
}

bool AtrMemoryImage::WriteBackImageToFile() const
{
	if (GetFilename()) {
		return WriteImageToFile(GetFilename());
	} else {
		return false;
	}
}

bool AtrMemoryImage::WriteImageToFile(const char* filename) const
{
	bool ret;
	EImageType imageType = DetermineImageTypeFromFilename(filename);

	if (imageType == eUnknownImageType) {
		DPRINTF("cannot determine image type from filename - using ATR format\n");
		imageType = eAtrImageType;
	}

#ifndef USE_ZLIB
	switch (imageType) {
	case eAtrGzImageType:
	case eXfdGzImageType:
	case eDcmGzImageType:
	case eDiGzImageType:
		DPRINTF("cannot write compressed file - zlib support disabled at compiletime!\n");
		return false;
	default: break;
	}
#endif

	switch (imageType) {
	case eAtrImageType:
	case eAtrGzImageType:
		ret = AtrMemoryImage::WriteImageToAtrFile(filename, imageType==eAtrGzImageType);
		break;
	case eXfdImageType:
	case eXfdGzImageType:
		ret = AtrMemoryImage::WriteImageToXfdFile(filename, imageType==eXfdGzImageType);
		break;
	case eDcmImageType:
	case eDcmGzImageType:
		ret = AtrMemoryImage::WriteImageToDcmFile(filename, imageType==eDcmGzImageType);
		break;
	case eDiImageType:
	case eDiGzImageType:
		ret = AtrMemoryImage::WriteImageToDiFile(filename, imageType==eDiGzImageType);
		break;
	case eUnknownImageType:
		DPRINTF("unknown image type!\n");
		return false;
	default:
		DPRINTF("unsupported image type!\n");
		return false;
	}
	return ret;
}

bool AtrMemoryImage::ReadImageFromAtrFile(const char* filename)
{
	unsigned char hdr[16];
	size_t imgSize, s;

	FileIO* fileio;

#ifdef USE_ZLIB
	fileio = new GZFileIO();
#else
	fileio = new StdFileIO();
#endif

	FreeImageData();

	if (!fileio->OpenRead(filename)) {
		DPRINTF("cannot open \"%s\" for reading\n",filename);
		return false;
	}

	if ( fileio->ReadBlock(hdr, 16) != 16 ) {
		DPRINTF("cannot read header\n");
		goto failure;
	}

	if (!SetFormatFromATRHeader(hdr)) {
		DPRINTF("SetFormatFromATRHeader failed\n");
		goto failure;
	}

	if ((imgSize = GetImageSize()) == 0) {
		DPRINTF("GetImageSize = 0\n");
		goto failure;
	}
	fData = new unsigned char[imgSize];
	memset(fData, 0, imgSize);

	if (!fData) {
		DPRINTF("cannot alloc fData\n");
		goto failure;
	}

	if ( (s=fileio->ReadBlock(fData, imgSize)) != imgSize) {
		DPRINTF("truncated ATR file: only got %d of %d bytes\n", s, imgSize);
		//goto failure;
	}

	fChanged = false;

	fileio->Close();
	delete fileio;
	return true;

failure:
	fChanged = false;
	FreeImageData();
	fileio->Close();
	delete fileio;
	return false;
}

bool AtrMemoryImage::WriteImageToAtrFile(const char* filename, bool useGz) const
{
	unsigned char hdr[16];
	size_t imgSize;

	FileIO* fileio;

	if (! fData) {
		DPRINTF("no fData\n");
		return false;
	}

	if (!CreateATRHeaderFromFormat(hdr)) {
		DPRINTF("CreateATRHeaderFromFormat failed\n");
		return false;
	}

	if ((imgSize = GetImageSize()) == 0) {
		DPRINTF("GetImageSize = 0\n");
		return false;
	}

	if (useGz) {
#ifdef USE_ZLIB
		fileio = new GZFileIO();
#else
		DPRINTF("cannot write compressed file - zlib support disabled at compiletime!\n");
		return false;
#endif
	} else {
		fileio = new StdFileIO();
	}

	if (!fileio->OpenWrite(filename)) {
		DPRINTF("cannot open \"%s\" for writing\n",filename);
		delete fileio;
		return false;
	}
	if (fileio->WriteBlock(hdr,16) != 16) {
		DPRINTF("cannot write ATR header\n");
		goto failure;
	}
	if (fileio->WriteBlock(fData,imgSize) != imgSize) {
		DPRINTF("cannot write ATR image\n");
		goto failure;
	}
	fChanged = false;
	fileio->Close();
	delete fileio;
	return true;

failure:
	fileio->Close();
	fileio->Unlink(filename);
	delete fileio;
	return false;
}

bool AtrMemoryImage::ReadImageFromXfdFile(const char* filename)
{
	size_t imgSize, s;
	int numSecs;
	ESectorLength seclen;

	FileIO* fileio;

#ifdef USE_ZLIB
	fileio = new GZFileIO();
#else
	fileio = new StdFileIO();
#endif

	FreeImageData();

	if (!fileio->OpenRead(filename)) {
		DPRINTF("cannot open \"%s\" for reading\n",filename);
		delete fileio;
		return false;
	}

	imgSize = fileio->GetFileLength();

	if ( (imgSize & 0x7f) || imgSize < 384 ) {
		DPRINTF("illegal image size %d\n", imgSize);
		goto failure;
	}

	if (imgSize & 0x80) {
		seclen = e256BytesPerSector;
		numSecs = (imgSize - 384) / 256 + 3;
	} else {
		seclen = e128BytesPerSector;
		numSecs = imgSize / 128;
	}
	SetFormat(seclen, numSecs);

	if (imgSize != GetImageSize()) {
		DPRINTF("setting image size failed!\n");
		goto failure;
	}
	fData = new unsigned char[imgSize];
	memset(fData, 0, imgSize);

	if (!fData) {
		DPRINTF("cannot alloc fData\n");
		goto failure;
	}

	if ( (s=fileio->ReadBlock(fData, imgSize)) != imgSize) {
		DPRINTF("truncated XFD file: only got %d of %d bytes\n", s, imgSize);
	}

	fChanged = false;

	fileio->Close();
	delete fileio;
	return true;

failure:
	fChanged = false;
	FreeImageData();
	fileio->Close();
	delete fileio;
	return false;
}

bool AtrMemoryImage::WriteImageToXfdFile(const char* filename, bool useGz) const
{
	size_t imgSize;
	FileIO* fileio;

	if (! fData) {
		DPRINTF("no fData\n");
		return false;
	}

	if ((imgSize = GetImageSize()) == 0) {
		DPRINTF("GetImageSize = 0\n");
		return false;
	}

	if (useGz) {
#ifdef USE_ZLIB
		fileio = new GZFileIO();
#else
		DPRINTF("cannot write compressed file - zlib support disabled at compiletime!\n");
		return false;
#endif
	} else {
		fileio = new StdFileIO();
	}

	if (!fileio->OpenWrite(filename)) {
		DPRINTF("cannot open \"%s\" for writing\n",filename);
		delete fileio;
		return false;
	}
	if (fileio->WriteBlock(fData,imgSize) != imgSize) {
		DPRINTF("cannot write XFD image\n");
		goto failure;
	}
	fChanged = false;
	fileio->Close();
	delete fileio;
	return true;

failure:
	fileio->Close();
	fileio->Unlink(filename);
	delete fileio;
	return false;
}

bool AtrMemoryImage::ReadImageFromDcmFile(const char* filename)
{
	DCMCodec* dcmcodec;
	FileIO* fileio;

#ifdef USE_ZLIB
	fileio = new GZFileIO();
#else
	fileio = new StdFileIO();
#endif
       	dcmcodec = new DCMCodec(fileio, this);

	FreeImageData();

	int ret;

	if ((ret=dcmcodec->Load(filename))) {
		fChanged = false;
	}

	delete dcmcodec;
	delete fileio;
	return ret;
}

bool AtrMemoryImage::WriteImageToDcmFile(const char* filename, bool useGz) const
{
	DCMCodec* dcmcodec;
	FileIO* fileio;
	if (useGz) {
#ifdef USE_ZLIB
		fileio = new GZFileIO();
#else
		DPRINTF("cannot write compressed file - zlib support disabled at compiletime!\n");
		return 1;
#endif
	} else {
		fileio = new StdFileIO();
	}

	dcmcodec = new DCMCodec(fileio, const_cast<AtrMemoryImage*>(this));

	bool ret;

	if ((ret=dcmcodec->Save(filename))) {
		fChanged = false;
	}

	delete dcmcodec;
	delete fileio;
	return ret;
}

unsigned char AtrMemoryImage::CalculateDiSectorChecksum(unsigned char* buf, size_t len) const
{
	size_t chksum = 0;
	size_t i;
	for (i=0;i<len;i++) {
		chksum += buf[i];
		if (chksum >= 256) {
			chksum -= 255;
		}
	}
	return chksum;
}

bool AtrMemoryImage::ReadImageFromDiFile(const char* filename)
{
	unsigned char buf[256];
	unsigned char* map;
	unsigned char unknown;
	size_t sectorsPerTrack;
        unsigned char tracksPerSide;
	unsigned char sides;
	size_t sectorLength;
	size_t totalSectors;
	ESectorLength density;

	size_t sector;

	FileIO* fileio;

#ifdef USE_ZLIB
	fileio = new GZFileIO();
#else
	fileio = new StdFileIO();
#endif

	FreeImageData();

	if (!fileio->OpenRead(filename)) {
		DPRINTF("cannot open \"%s\" for reading\n",filename);
		delete fileio;
		return false;
	}

	{
		// check header
		if (fileio->ReadBlock(buf,2) != 2) {
			goto failure_eof;
		}
		if (buf[0] != 'D' || buf[1] != 'I') {
			DPRINTF("not an DI file!\n");
			goto failure;
		}

		// check version
		unsigned int version;
		if (!fileio->ReadWord(version)) {
			goto failure_eof;
		}
		if (version != 0x0220) {
			DPRINTF("illegal version in DI image: %04x\n",version);
			goto failure;
		}

		// skip creator string
		bool ok;

		do {
			ok = fileio->ReadByte(buf[0]);
		} while (ok && buf[0]);

		if (!ok) {
			goto failure_eof;
		}
	}

	// read image header

	if (!fileio->ReadByte(unknown)) {
		goto failure_eof;
	}

	if (!fileio->ReadByte(tracksPerSide)) {
		goto failure_eof;
	}

	if (!fileio->ReadBigEndianWord(sectorsPerTrack)) {
		goto failure_eof;
	}

	if (!fileio->ReadByte(sides)) {
		goto failure_eof;
	}
	sides++;

	if (!fileio->ReadByte(unknown)) {
		goto failure_eof;
	}

	if (!fileio->ReadBigEndianWord(sectorLength)) {
		goto failure_eof;
	}

	if (!fileio->ReadByte(unknown)) {
		goto failure_eof;
	}

	totalSectors = sectorsPerTrack * tracksPerSide * sides;

	if (totalSectors == 0 || totalSectors > 65535) {
		DPRINTF("illegal number of sectors in DI image header: %d\n", totalSectors);
		goto failure;
	}

	switch (sectorLength) {
	case 128:
		density = e128BytesPerSector;
		break;
	case 256:
		density = e256BytesPerSector;
		break;
	default:
		DPRINTF("illegal sector length in DI image header: %d\n", sectorLength);
		goto failure;
	}

	if (!CreateImage(density, sectorsPerTrack, tracksPerSide, sides)) {
		goto failure;
	}

	// read sector (checksum) map
       	map = new unsigned char[totalSectors];

	if (fileio->ReadBlock(map, totalSectors) != totalSectors) {
		delete [] map;
		goto failure_eof;
	}

	for (sector=1; sector <= totalSectors; sector++) {
		unsigned char checksum = map[sector-1];

		if (checksum) {
			size_t len = GetSectorLength(sector);
			if (fileio->ReadBlock(buf, len) != len) {
				delete [] map;
				goto failure_eof;
			}
			if (CalculateDiSectorChecksum(buf, len) != checksum) {
				DPRINTF("checksum error in DI image at sector %d\n",sector);
				delete [] map;
				goto failure;
			}
			if (!WriteSector(sector, buf, len)) {
				DPRINTF("error storing sector %d in internal image!\n", sector);
				delete [] map;
				goto failure;
			}
		}
	}

	delete [] map;

	fChanged = false;

	fileio->Close();
	delete fileio;
	return true;

failure_eof:
	DPRINTF("error reading DI file: unexpected EOF\n");

failure:
	fChanged = false;
	FreeImageData();
	fileio->Close();
	delete fileio;
	return false;
}

bool AtrMemoryImage::WriteImageToDiFile(const char* filename, bool useGz) const
{
	size_t sectorsPerTrack;
        unsigned char tracksPerSide;
	unsigned char sides;
	size_t sectorLength;
	size_t totalSectors;
	size_t sector;
	unsigned char* map;
	unsigned char buf[256];

	FileIO* fileio;

	if (! fData) {
		DPRINTF("no fData\n");
		return false;
	}

	if ( GetSectorsPerTrack() < 1
	  || GetSectorsPerTrack() > 65535
	  || GetTracksPerSide() < 1
	  || GetTracksPerSide() > 255
	  || GetSides() < 1
	  || GetSides() > 255) {
		DPRINTF("image format incomatible with DI format!\n");
		return false;
	}
	sectorsPerTrack = GetSectorsPerTrack();
	tracksPerSide = GetTracksPerSide();
	sides = GetSides() - 1;
	totalSectors = GetNumberOfSectors();
	sectorLength = GetSectorLength();

	if (useGz) {
#ifdef USE_ZLIB
		fileio = new GZFileIO();
#else
		DPRINTF("cannot write compressed file - zlib support disabled at compiletime!\n");
		return false;
#endif
	} else {
		fileio = new StdFileIO();
	}

	if (!fileio->OpenWrite(filename)) {
		DPRINTF("cannot open \"%s\" for writing\n",filename);
		delete fileio;
		return false;
	}

	map = new unsigned char [totalSectors];

	// build sector map
	for (sector=1; sector<=totalSectors; sector++) {
		size_t len = GetSectorLength(sector);
		if (!ReadSector(sector, buf, len)) {
			DPRINTF("unable to read sector %d from memory image!\n",sector);
			goto failure;
		}
		map[sector-1] = CalculateDiSectorChecksum(buf, len);
	}

	// creator information
	strcpy((char*)buf,"AtariSIO");

	if ( !fileio->WriteByte('D')
	  || !fileio->WriteByte('I')
	  || !fileio->WriteWord(0x0220)
	  || fileio->WriteBlock(buf, strlen((char*)buf)+1) != strlen((char*)buf)+1
	  || !fileio->WriteByte(1)
	  || !fileio->WriteByte(tracksPerSide)
	  || !fileio->WriteBigEndianWord(sectorsPerTrack)
	  || !fileio->WriteByte(sides)
	  || !fileio->WriteByte(0)
	  || !fileio->WriteBigEndianWord(sectorLength)
	  || !fileio->WriteByte(0)
	  || fileio->WriteBlock(map, totalSectors) != totalSectors) {
		goto failure_eof;
	}

	// write non-zero sectors
	for (sector=1; sector<=totalSectors; sector++) {
		if (map[sector-1]) {
			size_t len = GetSectorLength(sector);
			if (!ReadSector(sector, buf, len)) {
				DPRINTF("unable to read sector %d from memory image!\n",sector);
				goto failure;
			}
			if (fileio->WriteBlock(buf, len) != len) {
				goto failure_eof;
			}
		}
	}

	delete [] map;
	fChanged = false;
	fileio->Close();
	delete fileio;
	return true;

failure_eof:
	DPRINTF("error writing DI file\n");
failure:
	delete [] map;
	fileio->Close();
	fileio->Unlink(filename);
	delete fileio;
	return false;
}

bool AtrMemoryImage::ReadSector(size_t sector,unsigned char* buffer,size_t buffer_length) const
{
	bool ret=true;
	int len, offset;

	if ((offset=CalculateOffset(sector)) < 0 ) {
		DPRINTF("illegal sector in ReadSector: %d\n", sector);
		return false;
	}

	if ((len=GetSectorLength(sector)) < 0) {
		DPRINTF("GetSectorLength(%d) failed\n",sector);
		return false;
	}

	if (!buffer_length) {
		DPRINTF("buffer length = 0\n");
		return false;
	}

	if (buffer_length < (size_t) len) {
		DPRINTF("buffer length < sector length [ %d < %d ]\n",buffer_length, (size_t) len);
		ret = false;
		len = buffer_length;
	} else if (buffer_length > (size_t)len) {
		DPRINTF("buffer length > sector length [ %d > %d ]\n",buffer_length, (size_t) len);
		ret = false;
	}

	memcpy(buffer, fData+offset, len);

	return ret;
}

bool AtrMemoryImage::WriteSector(size_t sector,const unsigned char* buffer,size_t buffer_length)
{
	int len, offset;

	if (IsWriteProtected()) {
		DPRINTF("attempting to write sector to write protected image\n");
		return false;
	}

	if ((offset=CalculateOffset(sector)) < 0 ) {
		DPRINTF("illegal sector in WriteSector: %d\n", sector);
		return false;
	}

	if ((len=GetSectorLength(sector)) < 0) {
		DPRINTF("GetSectorLength(%d) failed\n",sector);
		return false;
	}

	if (buffer_length != (size_t) len) {
		DPRINTF("buffer length = len [ %d != %d ]\n", buffer_length, (size_t) len);
		return false;
	}

	fChanged = true;
	memcpy(fData+offset, buffer, len);

	return true;
}

bool AtrMemoryImage::IsAtrMemoryImage() const
{
	return true;
}

void AtrMemoryImage::SetWriteProtect(bool on)
{
	//if (on != IsWriteProtected()) {
	//	fChanged = true;
	//}
	super::SetWriteProtect(on);
}

AtrMemoryImage::EImageType AtrMemoryImage::DetermineImageTypeFromFilename(const char* filename) const
{
	int len = strlen(filename);

	if (len < 3) {
		return eUnknownImageType;
	}
	if (strcasecmp(filename+len-3,".di") == 0) {
		return eDiImageType;
	}

	if (len < 4) {
		return eUnknownImageType;
	}
	if (strcasecmp(filename+len-4,".atr") == 0) {
		return eAtrImageType;
	}
	if (strcasecmp(filename+len-4,".xfd") == 0) {
		return eXfdImageType;
	}
	if (strcasecmp(filename+len-4,".dcm") == 0) {
		return eDcmImageType;
	}

	if (len < 6) {
		return eUnknownImageType;
	}
	if (strcasecmp(filename+len-6,".di.gz") == 0) {
		return eDiGzImageType;
	}

	if (len < 7) {
		return eUnknownImageType;
	}
	if (strcasecmp(filename+len-7,".atr.gz") == 0) {
		return eAtrGzImageType;
	}
	if (strcasecmp(filename+len-7,".xfd.gz") == 0) {
		return eXfdGzImageType;
	}
	if (strcasecmp(filename+len-7,".dcm.gz") == 0) {
		return eDcmGzImageType;
	}

	return eUnknownImageType;
}
