/*
   DCMCodec.C - DCM image format encoder/decoder

   Copyright (c) 2001-2003 Jindrich Kubec, Matthias Reichl

   The original code has been taken from Jindrich Kubec's 'Acvt' program
   and was adapted by Matthias Reichl for use with AtariSIO.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "DCMCodec.h"
#include <stdio.h>
#include <string.h>
#include "AtariDebug.h"

//#define DDPRINTF(x...) DPRINTF(x)
#define DDPRINTF(x...)

DCMCodec::DCMCodec(FileIO* ioclass, AtrMemoryImage* img)
	: fFileIO(ioclass),
	  fAtrMemoryImage(img)
{
}

DCMCodec::~DCMCodec()
{
}

bool DCMCodec::Load( const char* filename)
{
	unsigned char	btArcType = 0;		//Block type for first block
	unsigned char	btBlkType;		//Current block type

	fAlreadyFormatted = false;
	fLastPassFlag = false;
	fCurrentSector = 0;

	if (! fFileIO->OpenRead(filename)) {
		DPRINTF("cannot open \"%s\" for reading\n",filename);
		return false;
	}

	fFileLength = fFileIO->GetFileLength();

	for(;;) //outpass
	{
		if ( fFileIO->Tell() >= fFileLength )
		{
			if ( ( !fLastPassFlag ) && ( btArcType == eDCM_HEADER_MULTI ) )
			{
				DPRINTF("DCM: Multi-part archive error.\n" \
				"To process these files, you must first combine the files into a single file." );
				goto failure;
			}
		}

		if (!fFileIO->ReadByte(btArcType)) {
			goto failure_EOF;
		}

		switch( btArcType )
		{
			case eDCM_HEADER_MULTI:
			case eDCM_HEADER_SINGLE:
				if ( !DecodeRecFA() ) {
					DPRINTF("DCM: DecodeRecFA failed\n");
					goto failure;
				}
				break;

			default:
				DPRINTF("DCM: unknown header block 0x%02x.\n", btArcType );
				goto failure;
		}

		for(;;) //inpass
		{
			if (!fFileIO->ReadByte(btBlkType)) {
				goto failure_EOF;
			}

			if ( btBlkType == eDCM_PASS_END ) {
				DDPRINTF("pass end\n");
				break;
			}

			if ( fFileIO->Tell() >= fFileLength )
			{
				goto failure_EOF;
			}

			bool bRes = true;

			switch( btBlkType & 0x7F )
			{
				case eDCM_CHANGE_BEGIN:
					bRes = DecodeRec41();
					break;

				case eDCM_DOS_SECTOR:
					bRes = DecodeRec42();
					break;

				case eDCM_COMPRESSED:
					bRes = DecodeRec43();
					break;

				case eDCM_CHANGE_END:
					bRes = DecodeRec44();
					break;

				case eDCM_SAME_AS_BEFORE:
					//not needed
					//bRes = DecodeRec46();
					break;

				case eDCM_UNCOMPRESSED:
					bRes = DecodeRec47();
					break;

				default:
				{
					switch( btBlkType )
					{
						case eDCM_HEADER_MULTI:
						case eDCM_HEADER_SINGLE:
							DPRINTF("DCM: Trying to start section but last section never had "
							"an end section block.");
							break;

						default:
							DPRINTF("DCM: unknown block type 0x%02x - file may be corrupt.",btBlkType);
							break;
					}

					goto failure;
				}
			}

			if ( !bRes )
			{
				DPRINTF("DCM: Block %02X decode error!", btBlkType );
				goto failure;
			}

			size_t len = fAtrMemoryImage->GetSectorLength(fCurrentSector);

			bRes = fAtrMemoryImage->WriteSector(fCurrentSector, fCurrentBuffer, len);

			if (!bRes) {
				DPRINTF("DCM: writing sector %d to memory image failed!\n", fCurrentSector);
				goto failure;
			}

			if ( btBlkType & 0x80 ) {
				fCurrentSector++;
			} else {
				if (!fFileIO->ReadWord(fCurrentSector)) {
					goto failure_EOF;
				}
			}

		} //infinite for (inpass)

		//End block
		if ( fLastPassFlag ) {
			DDPRINTF("last pass - exit\n");
			break;
		}

	} //infinite for (outpass)

	fFileIO->Close();
	return true;

failure_EOF:
	DPRINTF("DCMCodec::Load(): unexpected EOF\n");
failure:
	fFileIO->Close();
	return false;

}

bool DCMCodec::DecodeRec41()
{
	DDPRINTF("%d: 41\n", fCurrentSector);
	unsigned char iOffset;
	unsigned char* pbt = fCurrentBuffer + iOffset;
       	if (!fFileIO->ReadByte(iOffset)) {
		goto failure_EOF;
	}
	pbt = fCurrentBuffer + iOffset;

	do
	{
		if (!fFileIO->ReadByte(*pbt)) {
			goto failure_EOF;
		}
		pbt--;
	} while( iOffset-- );

	return true;

failure_EOF:
	DPRINTF("DCMCodec::DecodeRec41(): unexpected EOF\n");
	return false;
}

bool DCMCodec::DecodeRec42()
{
	DDPRINTF("%d: 42\n", fCurrentSector);
	if (fFileIO->ReadBlock(fCurrentBuffer + 123, 5) != 5) {
		DPRINTF("DCMCodec::DecodeRec42(): unexpected EOF\n");
		return false;
	}
	memset( fCurrentBuffer, fCurrentBuffer[ 123 ], 123 );
	return true;
}

bool DCMCodec::DecodeRec43()
{
	DDPRINTF("%d: 43\n", fCurrentSector);
	unsigned char* pbtP = fCurrentBuffer;
	unsigned char* pbtE;

	unsigned char* pbtEnd = fCurrentBuffer + fSectorSize;

	size_t iTmp;
	unsigned char cTmp;

	do
	{
		//uncompressed string
		if ( pbtP != fCurrentBuffer ) {
			if (!ReadOffset(iTmp) ) {
				goto failure_EOF;
			}
			pbtE = fCurrentBuffer + iTmp;
		}
		else {
			if (! fFileIO->ReadByte(cTmp) ) {
				goto failure_EOF;
			}
			pbtE = fCurrentBuffer + cTmp;
		}

		if ( pbtE < pbtP )
			return false;

		if ( pbtE != pbtP )
		{
			if (fFileIO->ReadBlock(pbtP, pbtE - pbtP) != (size_t) (pbtE - pbtP)) {
				goto failure_EOF;
			}
			pbtP = pbtE;
		}

		if ( pbtP >= pbtEnd )
			break;

		//rle compressed string
		if (!ReadOffset(iTmp)) {
			goto failure_EOF;
		}
		pbtE = fCurrentBuffer + iTmp;

		if (!fFileIO->ReadByte(cTmp)) {
			goto failure_EOF;
		}

		if ( pbtE < pbtP )
			return false;

		memset( pbtP, cTmp, pbtE - pbtP );
		pbtP = pbtE;

	} while( pbtP < pbtEnd );

	return true;
failure_EOF:
	DPRINTF("DCMCodec::DecodeRec43(): unexpected EOF\n");
	return false;
}

bool DCMCodec::DecodeRec44()
{
	DDPRINTF("%d: 44\n", fCurrentSector);
	size_t iOffset;
	if (!ReadOffset(iOffset)) {
		goto failure_EOF;
	}

	if (fFileIO->ReadBlock( fCurrentBuffer + iOffset, fSectorSize - iOffset ) != (fSectorSize - iOffset) ) {
		goto failure_EOF;
	}

	return true;
failure_EOF:
	DPRINTF("DCMCodec::DecodeRec44(): unexpected EOF\n");
	return false;
}

bool DCMCodec::DecodeRec46()
{
	DDPRINTF("%d: 46\n", fCurrentSector);
	return true;
}

bool DCMCodec::DecodeRec47()
{
	DDPRINTF("%d: 47\n", fCurrentSector);
	size_t len;
        len = fSectorSize;

	if (fFileIO->ReadBlock(fCurrentBuffer, len) != len) {
		DPRINTF("DCMCodec::DecodeRec47(): unexpected EOF\n");
		return false;
	}
	return true;
}

bool DCMCodec::DecodeRecFA()
{
	unsigned char btPom;
	EDiskFormat dformat = eNoDisk;
	unsigned char btDensity;
	unsigned char btPass;
       
	if (!fFileIO->ReadByte(btPom)) {
		goto failure_EOF;
	}

	btDensity = ( btPom >> 5 ) & 0x03;
	btPass = btPom & 0x1F;
	fLastPassFlag = ( btPom & 0x80 ) ? true : false;

	DDPRINTF("%d: FA pass=%d last=%d\n", fCurrentSector, btPass, (int)fLastPassFlag);


	switch( btDensity )
	{
		case eDCM_DENSITY_SD:
			dformat = e90kDisk;
			break;

		case eDCM_DENSITY_ED:
			dformat = e130kDisk;
			break;

		case eDCM_DENSITY_DD:
			dformat = e180kDisk;
			break;

		default:
			DPRINTF("DCM: unknown density type 0x%02x\n", btDensity );
			return false;
	}

	if ( !fAlreadyFormatted )
	{
		if (!fAtrMemoryImage->CreateImage(dformat)) {
			DPRINTF("DCM: cannot format memory image!\n");
			return false;
		}
		fSectorSize = fAtrMemoryImage->GetSectorLength();

		fAlreadyFormatted = true;
	}
	
	if (!fFileIO->ReadWord(fCurrentSector)) {
		goto failure_EOF;
	}

	return true;
failure_EOF:
	DPRINTF("DCMCodec::DecodeRecFA(): unexpected EOF\n");
	return false;

}

bool DCMCodec::ReadOffset( size_t& off )
{
	unsigned char bt;
	if (!fFileIO->ReadByte(bt)) {
		DPRINTF("DCMCodec::ReadOffset(): unexpected EOF\n");
		return false;
	}
	if (bt == 0) {
		off=256;
	} else {
		off=bt;
	}
	return true;
}

bool DCMCodec::Save( const char* filename )
{
	int iDensity;
	fSectorSize = fAtrMemoryImage->GetSectorLength();

	switch (fAtrMemoryImage->GetDiskFormat()) {
	case e90kDisk:
		iDensity = eDCM_DENSITY_SD;
		break;
	case e130kDisk:
		iDensity = eDCM_DENSITY_ED;
		break;
	case e180kDisk:
		iDensity = eDCM_DENSITY_DD;
		break;
	default:
		DPRINTF("DCM: unsupported disk format!\n");
		return false;
	}

	if (!fFileIO->OpenWrite(filename)) {
		DPRINTF("cannot open \"%s\" for writing\n",filename);
		return false;
	}

	int iPass = 1;

	fPassBuffer = new unsigned char [ 0x6500 ];

	size_t iFirstSector = 0;
	size_t iPrevSector = 0;
	fCurrentSector = 1;
	size_t iNumSectors = fAtrMemoryImage->GetNumberOfSectors();

	memset( fPrevBuffer, 0, fSectorSize );
	memset( fCurrentBuffer, 0, fSectorSize );

	EncodeRecFA( false, iPass, iDensity, iFirstSector );

	//here should be other compression

	while( fCurrentSector <= iNumSectors )
	{
		iFirstSector = 0;

		while( ( fCurrentPtr - fPassBuffer ) < 0x5EFD )
		{
			if ( fCurrentSector > iNumSectors )
				break;

			size_t secLen = fAtrMemoryImage->GetSectorLength(fCurrentSector);

			if (!fAtrMemoryImage->ReadSector(fCurrentSector, fCurrentBuffer, secLen )) {
				DPRINTF("DCM: reading sector %d of memory image failed!\n",fCurrentSector);
				goto failure;
			}

			size_t i=0;
			while ( (i < secLen) && (fCurrentBuffer[i] == 0) ) {
				i++;
			}

			bool bSkip = (i==secLen);

			//first non empty sector is marked as first, what a surprise! :)
			if ( !bSkip && !iFirstSector )
			{
				iFirstSector = fCurrentSector;
				iPrevSector = fCurrentSector;
			}

			//if just skipped, increment sector
			if ( bSkip )
			{
				fCurrentSector++;
			}
			else
			{
				//if there is a gap, write sector number
				if ( ( fCurrentSector - iPrevSector ) > 1 )
				{
					*( fCurrentPtr++ ) = fCurrentSector;
					*( fCurrentPtr++ ) = fCurrentSector >> 8;
				}
				else
				{
					//else mark previous record
					*fLastRec |= 0x80;
				}

				//first sector could be encoded with only some data
				if ( fCurrentSector == iFirstSector )
					EncodeRec( true );
				else
				{
					//if are same, encode as record 46
					if ( !memcmp( fPrevBuffer, fCurrentBuffer, fSectorSize ) )
						EncodeRec46();
					else
						EncodeRec( false );
				}

				//store this sector as previous
				memcpy( fPrevBuffer, fCurrentBuffer, fSectorSize );

				//and move pointers
				iPrevSector = fCurrentSector;
				fCurrentSector++;
			}

		}

		*fLastRec |= 0x80;

		//encode end
		EncodeRec45();

		unsigned char* pEnd = fCurrentPtr;

		//change beginning block
		if ( fCurrentSector > iNumSectors )
			EncodeRecFA( true, iPass, iDensity, iFirstSector );
		else
			EncodeRecFA( false, iPass, iDensity, iFirstSector );

		//and write whole pass

		if ( ( pEnd - fPassBuffer ) > 0x6000 )
		{
			DPRINTF("DCM: internal error, pass too long!\n");
			goto failure;
		}

		if ( fFileIO->WriteBlock( fPassBuffer, pEnd - fPassBuffer ) != (size_t) (pEnd - fPassBuffer) )
		{
			DPRINTF("DCMCodec::Save(): error writing to file \"%s\"\n",filename);
			goto failure;
		}

		iPass++;
	}

	fFileIO->Close();
	delete [] fPassBuffer;
	return true;

failure:
	fFileIO->Close();
	fFileIO->Unlink(filename);
	delete [] fPassBuffer;
	return false;
}

void DCMCodec::EncodeRecFA( bool bLast, int iPass, int iDensity, int iFirstSec )
{
	DDPRINTF("%d: FA pass=%d last=%d\n", fCurrentSector, iPass, (int)bLast);
	fCurrentPtr = fPassBuffer;
	
	fLastRec = fCurrentPtr;

	unsigned char btType = bLast ? 0x80 : 0;

	btType |= ( iDensity & 3 ) << 5;

	btType |= ( iPass & 0x1F );

	*( fCurrentPtr++ ) = eDCM_HEADER_SINGLE;
	*( fCurrentPtr++ ) = btType;
	*( fCurrentPtr++ ) = iFirstSec;
	*( fCurrentPtr++ ) = iFirstSec >> 8;

}

void DCMCodec::EncodeRec45()
{
	DDPRINTF("%d: 45\n", fCurrentSector);

	fLastRec = fCurrentPtr;
	*( fCurrentPtr++ ) = eDCM_PASS_END;
}

void DCMCodec::EncodeRec46()
{
	DDPRINTF("%d: 46\n", fCurrentSector);
	
	fLastRec = fCurrentPtr;
	*( fCurrentPtr++ ) = eDCM_SAME_AS_BEFORE;
}

void DCMCodec::EncodeRec( bool bIsFirstSector )
{
	fLastRec = fCurrentPtr;

	unsigned char abtBuff41[ 0x300 ];
	unsigned char abtBuff43[ 0x300 ];
	unsigned char abtBuff44[ 0x300 ];
	unsigned char* abtBuff47 = fCurrentBuffer;

	int iEnd41 = 0x300;
	int iEnd43 = 0x300;
	int iEnd44 = 0x300;

	int iBestMethod = eDCM_UNCOMPRESSED;
	int iBestEnd = fSectorSize;
	unsigned char* pbtBest = abtBuff47;

	EncodeRec43( abtBuff43, &iEnd43, fCurrentBuffer, fSectorSize );

	if ( !bIsFirstSector )
	{
		EncodeRec41( abtBuff41, &iEnd41, fCurrentBuffer, fPrevBuffer, fSectorSize );
		EncodeRec44( abtBuff44, &iEnd44, fCurrentBuffer, fPrevBuffer, fSectorSize );
	}

	if ( iEnd41 < 255 && iEnd41 < iBestEnd )
	{
		iBestMethod = eDCM_CHANGE_BEGIN;
		iBestEnd = iEnd41;
		pbtBest = abtBuff41;
	}

	if ( iEnd43 < 255 && iEnd43 < iBestEnd )
	{
		iBestMethod = eDCM_COMPRESSED;
		iBestEnd = iEnd43;
		pbtBest = abtBuff43;
	}

	if ( iEnd44 < 255 && iEnd44 < iBestEnd )
	{
		iBestMethod = eDCM_CHANGE_END;
		iBestEnd = iEnd44;
		pbtBest = abtBuff44;
	}

	DDPRINTF("%d: %2x\n", fCurrentSector, iBestMethod);

	*( fCurrentPtr++ ) = iBestMethod;
	memcpy( fCurrentPtr, pbtBest, iBestEnd );
	fCurrentPtr += iBestEnd;
}

void DCMCodec::EncodeRec41( unsigned char* pbtDest, int* piDestLen, unsigned char* pbtSrc, unsigned char* pbtSrcOld, int iSrcLen )
{
	unsigned char* pbtS = pbtSrc + iSrcLen - 1;
	pbtSrcOld += iSrcLen - 1;

	unsigned char* pbtD = pbtDest;

	for( int i = 0; i < iSrcLen; i++ )
	{
		if ( *( pbtS-- ) != * ( pbtSrcOld-- ) )
			break;
	}

	pbtS++;

	*( pbtD++ ) = pbtS - pbtSrc;

	int iBytes = pbtS - pbtSrc + 1;

	while( iBytes-- )
	{
		*( pbtD++ ) = *( pbtS-- );
	}

	*piDestLen = pbtD - pbtDest;
}

void DCMCodec::EncodeRec43( unsigned char* pbtDest, int* piDestLen, unsigned char* pbtSrc, int iSrcLen )
{
	unsigned char* pbtEnd = pbtSrc + iSrcLen;
	unsigned char* pbtCur = pbtSrc;

	unsigned char* pbtD = pbtDest;

	while( pbtCur < pbtEnd )
	{
		bool bFound = false;

		for( unsigned char* pbtNow = pbtCur; pbtNow < ( pbtEnd - 2 ); pbtNow++ )
		{

			if ( ( *pbtNow == *(pbtNow+1) ) && ( *pbtNow == *(pbtNow+2) ) )
			{
				int iUnc = pbtNow - pbtCur;
				
				*( pbtD ++ ) = pbtNow - pbtSrc;
				if ( iUnc )
				{
					memcpy( pbtD, pbtCur, iUnc );
					pbtD += iUnc;
				}

				unsigned char bt = *pbtNow;
				unsigned char*p;
				for( p = pbtNow + 1; p < pbtEnd; p++ )
				{
					if ( *p != bt )
						break;
				}

				if ( p > pbtEnd )
					p = pbtEnd;

				*( pbtD++ ) = p - pbtSrc;
				*( pbtD++ ) = bt;

				pbtCur = p;
				bFound = true;
				break;
			}
		}

		if ( ( pbtCur >= pbtEnd - 2 ) || !bFound ) 
		{
			if ( pbtCur < pbtEnd )
			{
				*( pbtD++ ) = iSrcLen;
				memcpy( pbtD, pbtCur, pbtEnd - pbtCur );
				pbtD += pbtEnd - pbtCur;
			}

			break;
		}

	}

	*piDestLen = pbtD - pbtDest;
}

void DCMCodec::EncodeRec44( unsigned char* pbtDest, int* piDestLen, unsigned char* pbtSrc, unsigned char* pbtSrcOld, int iSrcLen )
{
	unsigned char* pbtS = pbtSrc;
	unsigned char* pbtEnd = pbtSrc + iSrcLen;

	unsigned char* pbtD = pbtDest;

	for( int i = 0; i < iSrcLen; i++ )
	{
		if ( *( pbtS++ ) != * ( pbtSrcOld++ ) )
			break;
	}

	pbtS--;

	*( pbtD++ ) = pbtS - pbtSrc;
	memcpy( pbtD, pbtS, pbtEnd - pbtS );
	pbtD += pbtEnd - pbtS;

	*piDestLen = pbtD - pbtDest;
}

